/* @(#) samples/c/amqsblst.c, pcode, p520, p520-aux-CSD05G  1.19.1.1 00/11/13 16:36:47 */
/* Simple MQ bandwidth and connectivity tester.
 *
 *      Pyramid Technology Corporation - Copyright (c) 1995, 1998.
 *      All Rights Reserved
 *
 *      THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF Pyramid.
 *      The copyright notice above does not evidence any
 *      actual or intended publication of such source code.
 *
 * This program is provided in the hope that it will be useful, but comes
 * with no guarantees whatsoever and no responsibility or liability is
 * accepted by the author or by Pyramid Technology Corporation for any loss
 * or consequential loss that may be suffered by the use (or otherwise) of
 * this program or of any program derived from it. Comments elsewhere in
 * this program must not be interpretted as varying the statement in this
 * paragraph.
 *
 * Compile & run me for a usage message.
 *      blast -x
 * will generate an explanation/usage message.
 *
 * One invocation can be used to send messages, the other to read them.
 *
 * This does not do as much validation of command line arguments as might be
 * desired in an industrial strength program.
 *
 * Time intervals, in many places, is stored in miliseconds (ms). This is OK
 * unless you plan to a single run of more than 3 weeks or so (32 bit M/C).
 *
 * Examples:
 *      blast QueueManager Queue
 * will write 1000 1K messages into Queue managed by QueueManager. The messages
 * will contain a sequence number.
 *
 *      blast -r QueueManager Queue
 * will read the messages out of the queue. Blast will check the message
 * sequence number, ie it is expected that the messages were generated
 * by blast. This checking may be suppressed with the -C option:
 *      blast -r -C QueueManager Queue
 *
 * The CRC after the message header is checked on message that are read
 * as long as the message is at least MIN_MESSAGE_SIZE (60) bytes long.
 * If the message is read on a machine which has a different CCSID than from
 * where it was sent, it is possible that the CRC would be invalid (if one
 * was ASCII and the other EBCDIC this would certainly be true). The CRC is
 * represented as to hex digits to give it the best chance of not being
 * corrupted by code conversion. The CRC is not checked if the -C option is
 * given (in which case nothing is checked), the CRC check alone may be
 * switched off with the -X option:
 * blast -r -X QueueManager Queue
 *
 * Messages contain the text 'M sss nnnn rrrr hh:mm:ss secs.msec', where 'M'
 * is replaced by a 'L' for the last message, sss is the message sequence
 * (starting from 1), nnnn is the number of messages that are to be sent,
 * rrrr is the seed for the random number generator, hh:mm:ss is the local
 * time, secs is the time in seconds since the epoch, and msec is
 * miliseconds. This is followed by the NUL byte, a 2 byte CRC and then a
 * repeating pattern of alphanumerics to fill the buffer.
 * The CRC is an XOR of all the bytes up to the NUL byte. The CRC is not
 * checked if -C is given, nor is it recalculated in an echo if -C is given.
 * Blast will check for a 'L' in messages that it reads and stop when it finds
 * one (but see -C and -L).
 * Messages are marked as being strings - which can thus be translated into
 * a different codeset for the receiving application.
 *
 * By default the sequence number starts from 1, this may be changed with -N:
 *      blast    -N 21 QueueManager Queue
 *      blast -r -N 21 QueueManager Queue
 * the first message sent/read should have a sequence number of 21. Note that
 * the end sequence remains unadjusted at the number of messages to be sent.
 *
 * If you wish to see the contents of messages received by blast, use -D nn,
 * where nn is the maximum number of bytes to be displayed. If nn is -1 the
 * whole message will be displayed. -D 0 displays nothing. Eg:
 *      blast -r -D 200 QueueManager Queue
 *
 * Message header information can be shown with -H nn, where nn is a number
 * (0..5) indicating how much information will be shown. 0 is nothing, 5 is
 * all of the header, in between the choice of what is included is arbitrary.
 *      blast -r -H 3 QueueManager Queue
 *
 * Blast will stop if it finds a 'Last' message. This enables a
 * pair of blasts to close a link nicely. Blast won't do this if -C is set.
 * You should arrange that it terminates on message count (-c) or on a timeout
 * (-G).
 * If -L n is blast will stop after receiving n 'Last' messages - presumably
 * from n corespondents/clients. If 'n' is -1 ignore all 'Last' messages.
 *
 * The number of messages to be send/received may be set with -c:
 *      blast -c 2000 QueueManager Queue
 *
 * If the messages received come from another blast, the number of
 * messages may be learned from the sending blast (see more later):
 *      blast -r -l QueueManager Queue
 *
 *      blast -t 9 QueueManager Queue
 * will tell you exactly what it is doing - as it happens. Unless you are
 * debugging a trace level of 1 or 2 is what you want.
 *
 *      blast -r -t 2 -s 100 QueueManager Queue
 * will read 1000 messages, and give a timing summary every 100 messages and
 * an overall summary. Having intermediate summaries is useful as it shows
 * you how things change as machine workload changes, etc.
 *
 *      blast -r -t 2 -p 10 QueueManager Queue
 * will read 1000 messages, and give a progress message every 10 messages. The
 * progress message is time stamped.
 *
 *      blast -r -t 2 -p 10 -d -s 100 QueueManager Queue
 * will read 1000 messages, and give a progress message every 10 messages. The
 * progress message is just a dot. Every 100 messages a timing summary is
 * printed.
 *
 *      blast -r -t 2 -d -p 10 -s 100 QueueManager Queue &
 *      sleep 10
 *      blast    -t 1                 QueueManager Queue
 * The first reads 1000 messages from the queue, it prints a dot every 10
 * messages, and a timing summary every 100 messages.
 * The sleep is to allow the first blast to get established before the second
 * blast is run to put the 1000 messages into the queue.
 *
 *      blast -r -t 2 -d -p 10 -s 100 -c -1 QueueManager Queue &
 *      sleep 10
 *      blast    -t 1                 -c -1 QueueManager Queue
 * This is as the previous example, but it will continue forever - or until you
 * press interrupt (control C) at the keyboard.
 *
 *      blast -r -t 2 -d -p 10 -s 100 -c -1 -e QueueManager Queue1 Queue2 &
 *      sleep 10
 *      blast    -t 1                 -c -1 -e QueueManager Queue1 Queue2
 * The second blast sends a message to Queue1, the first blast read it from
 * Queue1 and writes a reply to Queue2, the second blast reads the reply from
 * Queue2 and writes another message to Queue1. This will continue until
 * either of the blasts is interrupted. One of the main motivations of this
 * test is that it will not run the risk of filling a queue up and so can be
 * left to run for a long time.
 * It can be illustrative to run this between queue managers on different
 * machines - ie each blast runs on a different box. In this case both blasts
 * should trace progress (-t2 -d -p 10 -s 100).
 *
 * With the use of -e as above please note that: 1) blast spends most of
 * it's time waiting for a message to arrive from it's peer (and so doesn't
 * put a real load onto the system), 2) the same volume of messages flows in
 * each direction. It these are undesirable the -F option may be of use:
 * -F FlowInterval, FlowInterval is a number and the reader/echoer will
 * send one synchronisation reply when it receives the first message and one
 * for every FlowInterval messages that it receives.
 * The writer will expect one reply after it has sent a multiple of
 * FlowInterval messages. This allows high volume without running the
 * risk of over filling queues. Because the reader echoes the first
 * message both blasts should run without any pauses.
 * The -F option implies the -e option:
 *      blast -r -F 100 QueueManager Queue1 Queue2 &
 *      blast    -F 100 QueueManager Queue1 Queue2
 *
 * When reading from a queue blast will wait indefinitely for messages to
 * arrive. Should there be something wrong, this can mean that a test script
 * could run forever. This is where the -G option is useful:
 *      blast -r -G 100000 QueueManager Queue
 * will timeout if blast has to wait more than 100 seconds to receive any
 * message (including the first). Upon timeout blast will fail.
 *
 * Blast can be made to operate as a echo server, sending messages onwards to
 * many different queues using the Reply_to_Queue in the message header. Blast
 * will only talk to one queue manager. The -E option is used to specify this:
 *      blast -r -t 1 -c -1 -G 30000 -C -e -E 2 QueueManager Queue1 &
 *      sleep 10
 *      blast    -t 1 -e -E Queue2 QueueManager Queue1 Queue2 &
 *      blast    -t 1 -e -E Queue3 QueueManager Queue1 Queue3
 * The first blast is the echo server. It will echo messages to the reply queue
 * specified in the message header. It must not check message contents as the
 * sequence numbers will be mixed up from it's two clients. It will time out
 * after 30 seconds of waiting for a message.
 * The two other blasts are clients, they send to Queue1 and read from Queue2
 * or Queue3 respectively, they each send 1000 messages (the default).
 * Note how the argument to the -E option is different for the client and server.
 * For the server the argument is how many reply queues to keep open (for
 * efficiency) if it has to reply to more than this it will close/open.
 * For the client the argument to -E is the name of the reply queue that is
 * put into the message header, this may be different from the queue name after
 * the queue manager name - with interesting effects, but beware filling queues.
 *
 * Sending different sized messages can be a better test of a queue manager
 * than messages all the same size. The -R option will do this. If an error is
 * found it is imperative that the sequence can be reproduced, thus the seed
 * for the random number generator is always printed out when -R is used. To
 * reproduce the sequence set the seed with the -S option.
 *      blast -R -S 124189542 QueueManager Queue
 * Messages will be up to the maximum message size with a minimum of
 * MIN_MESSAGE_SIZE (60) bytes - unless the -C option is specified in
 * which case messages can be of 0 length.
 * A reader will not check the contents of a message if it is not at least
 * MIN_MESSAGE_SIZE bytes long. The first message is always at least that long
 * so that important numbers (like the seed) can get transmitted to the reading
 * blast.
 *
 * Blast checks that the size of message received is the Buffer size (-b).
 * Blast does not check received message sizes if random is set (-R)
 * unless the seed is also given (-S).
 *      blast -r -R -S 124189542 QueueMananger Queue
 * Beware the sequence of random numbers (generated by rand()) may not be
 * the same on different machine types given the same seed. This means that
 * checking of message size may give lots of errors, see -M below.
 * See also the '-l' option described below.
 *
 * Blast will not check the message size if -M is specified. This is useful
 * where a blast has several clients which send it randomly sized messages:
 *      blast -r -t 1 -c -1 -G 30000 -b 100000 -C -M -e -E 2 QueueManager \
 *                                                              Queue1 &
 *      sleep 10
 *      blast    -t 1 -R -b 100000 -e -E Queue2 QueueManager Queue1 Queue2 &
 *      blast    -t 1 -R -b 10000  -e -E Queue3 QueueManager Queue1 Queue3
 * Note that the client buffer sizes don't need to be the same, but that the
 * server must have a buffer size that is at least as large as the largest.
 *
 * If variable message sizes are expected -R may be used to receive any
 * message up to the buffer size:
 *      blast -r -R QueueMananger Queue
 * will receive 1000 messages up to 1K in length.
 *
 * It may be inconvenient to tell the receiving blast how many messages it is
 * to receive and what the random message length seed is; it picks these up
 * from the first message. The -l option may be used to tell it to 'learn'
 * these numbers from the sending blast. Note this implies that the receiving
 * blast has to inspect the message contents, it will do this if -l is set
 * on the first message even if -C is set. The length of the first message will
 * always be at least MIN_MESSAGE_SIZE.
 *
 * Blast will set a message priority when sending message if given -P, this
 * must be a number in the range 0..9. Eg:
 *      blast -P 8 QueueMananger Queue
 * On reading this has no effect unless the messages are echoed. A special
 * value of -1 means echo the message with the priority that it was read. Eg:
 *      blast -r -e -P -1 QueueMananger Queue1 Queue2
 *
 * A Correlation Id may be specified with -I text, this will be inserted on
 * messages that are sent, and will select messages read from a queue.
 * Blast will copy the correlation id on messages that are echoed (-e).
 * This mechanism may be used to have one echo server service several clients,
 * all using the same queues, each of which only sees it's messages.
 *      blast -r -c -1 -e -L 2 -C QueueManager Queue1 Queue2 &
 *      blast -I one -e           QueueManager Queue1 Queue2 &
 *      blast -I two -e           QueueManager Queue1 Queue2
 *
 * Due to `end effects' blast can compete for system resources with channel
 * programs if it were to send another message immediately after sending one
 * (or reading one). To prevent this the -W option can specify a pause once
 * per iteration. In a writer this will be before the mqput, in a reader just
 * before the mqget.
 * The timings (put/get) will not include the time asleep, summaries of time
 * asleep will be produced, note that the first interval will appear to sleep
 * less because a sleep is not done before the very first get/put.
 * Thus to have a two second delay on putting messages into a queue:
 *      blast -W 2 QueueManager Queue1
 *
 * It is possible to set Message Descriptor Report Options on a message.
 * These tell MQ what to do when a message gets to the other end.
 * See the MQSeries Application Programming Reference for all the
 * gory details - Chapter 1 (Data type descriptions), under message
 * descriptor (MQMD).
 * You do this with the -o option, eg:
 *      blast -o MQRO_DEAD_LETTER_Q -o MQRO_COA -E Queue2 QueueManager Queue1
 * will send messages with the reply queue set as Queue2, one message will
 * be put on Queue2 for every message put on Queue1 (this is what MQRO_COA
 * means). If it can't be put on Queue1 the message will be put on the dead
 * letter queue (as per MQRO_DEAD_LETTER_Q).
 * Note that this will leave many messages on Queue2 - clear them out!!
 * Note that -o options are read in order on the command line and are additive,
 * but that MQRO_NONE will unset any options so far specified. The initial
 * value is MQRO_NONE.
 * The values are sent back in the Feedback field of the message descriptor
 * of the report message, you may see these with the -H option:
 *      blast -r -H 3 QueueManager Queue2
 *
 * By default all operations (get and put) are done outside of a unit of work
 * (NO_SYNCPOINT). This means that the effects of the operation is immediately
 * visible to other applications. You can specify that operations will be
 * grouped into units of work with the -u number_of_operations_in_unit_of_work
 * option:
 *      blast -u 10 QueueManager Queue
 * will put 1000 messages in 100 units of work. No special action is taken if
 * the number of messages is not exactly divisible by the number of operations
 * in a unit of work; this will result in the operations being rolled back when
 * blast exits (which may result in the message with the Last_Message marker
 * being lost).
 * If the echo (-e) or flow (-F) option is used the echoed messages will also
 * be got/put in units of work; these are the same units of work.
 *
 * You can arrange for some units of work to be backed out (rolled back):
 *      blast -u 10 -U 3 QueueManager Queue1
 * will result in every 3rd unit of work being cancelled (ie messages 21 to
 * 30 will be lost - which may upset a blast which reads from Queue1). -U is
 * ignored if -u is not specified.
 *
 * Bugs:
 * -    If -w is given messages will be left in the queue.
 * -    The examples above assume that you are running a UNIX shell.
 * Beware:
 * -    Timings are calculated in different ways, some are driven off the time
 *      of the last message, others accumulated times. This can lead to small
 *      discrepancies between what should be the same numbers.
 *
 * Assume that we are compiling for UNIX unless we know better. Appropriate
 * defines are:
 * AMQ_OS2      OS/2
 * AMQ_NT       NT
 * AS400        OS/400 on AS/400
 * MVS          IBM Mainframe
 * AMQ_UNIX     UNIX - but that is default anyway
 */

/* What are we compiling for ? */
#ifndef AMQ_OS2
#ifndef AMQ_NT
#ifndef MVS
#ifndef AMQ_UNIX
#define AMQ_UNIX        1       /* A good choice */
#endif
#endif
#endif
#endif

#include <cmqc.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <signal.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h>

/* We need to have a struct that provides subsecond timers.
 * Guaranteed different for every operating system.
*/
#ifdef  AMQ_UNIX
#include <sys/time.h>
typedef struct  timeval tval_t;         /* For convenience */

#endif   /* AMQ_UNIX */

#ifdef  AMQ_OS2
#define INCL_DOSDATETIME
#define  INCL_DOSPROCESS         /* To get DosSleep() prototype */
#include <os2.h>
#include <process.h>

/* The rest of the program expects to see something like this */
typedef struct  {
   long    tv_sec;
   long    tv_usec;
} tval_t;

/* We use getpid as a seed for random numbers */
#define getpid()        _getpid()
#define sleep(tim)      DosSleep(tim / 1000)

#endif   /* AMQ_OS2 */

#ifdef  AMQ_NT
#include <windows.h>

typedef struct  {
   long    tv_sec;
   long    tv_usec;
} tval_t;

/* We use getpid as a seed for random numbers */
#include <process.h>
#define getpid()        _getpid()
#define sleep(tim)      Sleep(tim)

#endif   /* AMQ_NT */

#define Kilo            1024
#define DefBufferSize   Kilo
#define DefNumMessages  1000
#define DefProgCnt      100
#define MIN_MESSAGE_SIZE 60             /* Min message size */
#define HDR_DUMP_MAX    5               /* Max value of HdrDump */
#define  CRC_BYTES   2     /* # bytes that make up the CRC */

/* Trace levels */
#define ProgLev                 2       /* Produce progress if Trace >= this */
#define DefTrace                0
#define TR_SHOW_CONTENTS        3       /* Show message contents */

/* An Echo queue buffer.
 * This structure buffers open handles to echo queues so that we don't have
 * to keep on opening them.
 * This is stored as an array, an unused slot is indicated by ec_used == 0.
 */
typedef struct  {
   MQCHAR48        ec_qname;       /* Name of the queue */
   MQHOBJ          ec_qmd;         /* Queue handle */
   long            ec_used;        /* When last used- ms since start */
} Ecb;

int     BufferSize = DefBufferSize;     /* How much to send */
long    NumMessages = DefNumMessages;   /* How many to send */
int     Trace = DefTrace;               /* Trace level */
long    DumpInMessage;                  /* Dump incoming msg chrs -1 is all */
int     HdrDump;                        /* Level of header info to dump */
int     ProgCnt = DefProgCnt;           /* Progress count */
int     ProgDot;                        /* Print a dot as a progress message */
int     ProgMsg;                        /* Msgs since progress report */
int     ReadFromQueue;                  /* True if to read */
char*   CorrelId;                       /* Correlation Id or NULL */
char*   Message;                        /* Buffer - size BufferSize */
char*   progname;                       /* Who we are */
char*   QueueManager;                   /* Queue Manager */
char*   QueueName;                      /* Queue Name */
char*   EchoQueueName;                  /* Queue Name for echoing - unless -E */
tval_t  ProgStart;                      /* When the program started */
tval_t  now;                            /* Time now */
int     WriteWait;                      /* Fill Q with this before reading */
int     GotSig;                         /* Caught interrupt */
int     IgnoreLast;                     /* # 'Last' messages to ignore */
int     echo;                           /* Echo option */
int     FlowInterval;                   /* Flow control interval */
int     flush;                          /* Flush the read queue - echo mode */
int     exit_c;                         /* Code to exit with */
int     ContentCheck = 1;               /* Check message contents are as expected */
long    SeqStart;                       /* Sequence number start */
int     SetParamsFromFirstMessage;      /* On reader side */

MQLONG  ReportOption = MQRO_NONE;       /* May get set by the user */

int     PrioritySet;                    /* MsgPriority has a value */
int     MsgPriority;                    /* Message priority. If reading/echoing and
                                    * this is -1, echo with the priority read.
                                    */

/* Stats gathering */
long    Stats;                          /* Print intermediate stats if != 0 */
tval_t  IntervalStart;                  /* Time of 1st msg in interval */
long    MaxInterval = -1;               /* Longest Interval - m seconds */
long    MinInterval = -1;               /* Shortest Interval - m seconds */
long    TotInterval;                    /* Length of measured intervals - ms */
long    NumIntervals;                   /* How many intervals there were */
long    TotMsgs;                        /* # measured messages */
int     MsgNo;                          /* As recorded by PrintTrace() */
double  BytesInterval;                  /* Bytes sent in interval */
double  TotBytesSent;                   /* Total bytes sent/received */

/* Things needed for random size message processing.
 * SizeBuf is only needed if -w (WriteWait) is specified with echo.
 */
int     RandSize;                       /* True if send random sized messages */
unsigned int    seed;                   /* Random number Seed */
int     SeedGiven;                      /* Seed given as an option */
long    LastMsgSize;                    /* Last message size generated */
long*   SizeBuff;                       /* Circular buffer to hold msg sizes */
long*   WriteSize;                      /* Msg just sent */
long*   ReadSize;                       /* Next msg to read */
int     DontCheckSize;                  /* Don't check size of msgs got */

/* NoWait (-n) option */
int     NoWait;                         /* Don't wait on reads */
long    WaitTime;                       /* Time spent waiting this interval */
long    MaxWaitTime = -1;               /* Longest - any interval */
long    MinWaitTime = -1;               /* Shortest - any interval */
long    TotWaitTime;                    /* Time spent waiting - total */
MQGMO   DefGmo = { MQGMO_DEFAULT };     /* For resetting to nowait */

MQLONG  GetTimeout;                     /* How long to wait for messages */

/* Sleep (-W) option */
int     LoopWait;                       /* Delay at top of each loop */
long    TimeLoopWait;                   /* Time this interval sleeping */
long    TotTimeLoopWait;                /* Total time sleeping */

/* Echo Server/Client options -E */
char*   EchoServerOption;               /* -E option */
int     EchoQueuesOpen;                 /* # Echo queues */
Ecb*    EchoHandles;                    /* Vector of echo queue handles */

/* Units of work (-u and -U), syncpoints */
int     UnitOfWork;                     /* Operations in a Unit of Work */
int     UnitsBackedOut;                 /* Back out 1 in this UOWs */
int     UwCount;                        /* Operations so far in UOW */
int     UwUnits;                        /* Units of work so far */

/* Fill the message with a repeating pattern of alphanumerics.
 * Below is what we use - should be possible to convert to any other character
 * set.
 */
int   CheckCRC = 1;        /* Please check the CRC on msgs */
char  MsgTxt[] = "abcdefghijklmnopqrstuvwxyz0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
char  HexDigits[] = "0123456789abcdef";

/* We can't use this constant as a literal */
char    mqfmt_string_array[] = { MQFMT_STRING_ARRAY };

/* Something to hold the names of the report options.
 * We know that we have reached the end by the NULL pointer.
 * Note that MQRO_NONE is handled specially as it resets the value to 0,
 * all the others are additive (ored).
 */
typedef struct  {
   MQLONG  ro_val;                 /* What value to OR in */
   char*   ro_name;                /* What the user types */
} Repo;

Repo    RepoList[] = {
   MQRO_EXCEPTION,                 "MQRO_EXCEPTION",
   MQRO_EXCEPTION_WITH_DATA,       "MQRO_EXCEPTION_WITH_DATA",
   MQRO_EXCEPTION_WITH_FULL_DATA,  "MQRO_EXCEPTION_WITH_FULL_DATA",
   MQRO_EXPIRATION,                "MQRO_EXPIRATION",
   MQRO_EXPIRATION_WITH_DATA,      "MQRO_EXPIRATION_WITH_DATA",
   MQRO_EXPIRATION_WITH_FULL_DATA, "MQRO_EXPIRATION_WITH_FULL_DATA",
   MQRO_COA,                       "MQRO_COA",
   MQRO_COA_WITH_DATA,             "MQRO_COA_WITH_DATA",
   MQRO_COA_WITH_FULL_DATA,        "MQRO_COA_WITH_FULL_DATA",
   MQRO_COD,                       "MQRO_COD",
   MQRO_COD_WITH_DATA,             "MQRO_COD_WITH_DATA",
   MQRO_COD_WITH_FULL_DATA,        "MQRO_COD_WITH_FULL_DATA",
   MQRO_COPY_MSG_ID_TO_CORREL_ID,  "MQRO_COPY_MSG_ID_TO_CORREL_ID",
   MQRO_PASS_CORREL_ID,            "MQRO_PASS_CORREL_ID",
   MQRO_NEW_MSG_ID,                "MQRO_NEW_MSG_ID",
   MQRO_PASS_MSG_ID,               "MQRO_PASS_MSG_ID",
   MQRO_DEAD_LETTER_Q,             "MQRO_DEAD_LETTER_Q",
   MQRO_DISCARD_MSG,               "MQRO_DISCARD_MSG",
   (MQLONG)0,                      (char*)NULL
};

#ifdef AMQ_PC
/* We can not use xugGetOpt () as it is case insensitive. We must
 *provide our own version
 */
int     getopt(int argc, char **argv, char *opts);
int     optind = 1;
int     optopt;
char    *optarg;
#else
extern  int     optind;
extern  char*   optarg;
#endif

void    Allocate(void);
char  CalculateCRC(char* Msg);
void    CatchIntr(int sig);
void    CheckContents(char* Mbuf, long MsgLen, int sequence, char* what);
void  CheckMsgAlphaNumerics(char* Mbuf, long MsgLen, int sequence, char* what);
void    CheckUnitOfWork(void);
void    DoReads(void);
void    DoWrites(void);
void    DumpHeader(MQMD* InQmd);
void    DumpStr(char* str, int len);
void    DumpText(char* Message, MQLONG Len);
long    ExpectedMsgSize(void);
char*   FormatNow(void);
int     FlushEchoQueue(void);
void    GetArgs(int argc, char* argv[]);
void    Init(void);
void    Init2(void);
void  InsertCRC(char* Msg);
int     IsZeroTime(tval_t tp);
long    MsDiff(tval_t start, tval_t end);
long    MsgSize(void);
char*   OpenEchoQueue(MQMD* Mqmd, MQHOBJ* obj);
char*   OpenQueue(int WriteFlag, MQHOBJ* hObjp, int EchoQueue);
int     ParseReportOption(char* opt);
char*   PrintAve(float Num1, long Num2);
void    PrintStats(void);
void    PrintTrace(char* dirn, char* fmt, ...);
long    RandMsgSize(void);
void    SetNow(void);
void    SleepABit(void);
char*   strmake(char* str);
void    TimeMessage(char* fmt, ...);
void    Usage(int ec);
void    ZeroTime(tval_t* tp);

MQHCONN hQm          = MQHC_UNUSABLE_HCONN;

int main(int argc, char * argv[])
{
   MQLONG  CompCode, Reason;

   Init();

   GetArgs(argc, argv);

   Init2();

   /* Go until get interrupt ? */
   if(NumMessages == -1)
           signal(SIGINT, CatchIntr);

   /* Connect to Queue Manager
    */
   MQCONN(QueueManager, &hQm, &CompCode, &Reason);
   if(Trace || Reason != MQCC_OK)
           printf("MQCONN Compcode(%ld) Reason(%ld)\n", CompCode, Reason);

   /* Did that work ? */
   if(Reason == MQCC_OK) {
           if(ReadFromQueue)
                   DoReads();
           else
                   DoWrites();

           /* Disconnect
            */
           MQDISC(&hQm, &CompCode, &Reason);
           if(Trace)
                   printf("\nMQDISC Compcode(%ld) Reason(%ld)\n",
                           CompCode, Reason);
   }
   else
           exit_c = 2;

   PrintStats();

   return exit_c;
}

/* Pre argument parsing initialisation.
 */
void    Init(void)
{
   long    l;

   /* Set random number seed, good enough but not good.
    * A seed of '0' means unset - our own convention.
    */
   while(seed == 0) {
           SetNow();
           l = (long)(now.tv_sec) ^ (long)(now.tv_usec);
           l ^= ((long)getpid() << 8);
#ifdef  AMQ_UNIX
           l ^= (long)getppid() << 16;
#endif
           seed = (unsigned int)l;
   }

   /* I don't trust compilers to initialise these properly */
   BytesInterval = TotBytesSent = 0.0;

   /* We need to be able to compute time since the program started */
   SetNow();
   ProgStart = now;
}

/* Post argument parsing initialisation.
 */
void    Init2(void)
{
   /* If random message size *always* output the seed.
    * If we don't and we get an error we may not be able to reproduce
    * the error.
    * Note that we may not yet know what the seed is.
    */
   if(RandSize && !SetParamsFromFirstMessage) {
           printf("%s: Seed for random message size: %u\n", progname,
                   seed);
           srand(seed);
   }

   /* Make buffers */
   Allocate();

   /* If we are sending random sized messages, but delay before reading
    * them, we need to record how big they are so that we can check
    * when we do get round to reading them.
    */
   if(RandSize && WriteWait != 0 && !ReadFromQueue && echo) {
           if((SizeBuff = (long*)calloc((size_t)WriteWait + 1,
                           sizeof(long))) == (long*)NULL) {
                   fprintf(stderr, "%s: Can't allocate %d longs\n",
                           progname, WriteWait);
                   exit(2);
           }
           WriteSize = ReadSize = SizeBuff;
   }

   /* Do we need to be able to buffer a set of echo queue handles ? */
   if(ReadFromQueue && EchoServerOption != (char*)NULL) {
           if((EchoHandles = (Ecb*)calloc((size_t)EchoQueuesOpen,
                                   sizeof(Ecb))) == (Ecb*)NULL) {
                   fprintf(stderr,
                           "%s: Can't allocate Echo Queue buffers\n",
                           progname);
                   exit(2);
           }
   }
}

/* **** Command line argument parsing & Usage message ****
 *
 * Read command line arguments.
 */
void    GetArgs(int argc, char* argv[])
{
   int     o;
   int     ec;             /* Exit code - if we do */
   long    play_l;
   int     EchoServer;

   ec = 0;

   /* What are we called ? */
   if((progname = strrchr(argv[0], '/')) != (char*)NULL)
           progname++;
   else
           progname = argv[0];

   while((o = getopt(argc, argv,
           "b:c:CdD:eE:fF:G:H:I:lL:MnN:o:p:P:rRs:S:t:u:U:w:W:xX")) != EOF) {
           switch(o) {
           case 'b':       /* Buffer size */
                   if(sscanf(optarg, "%d", &BufferSize) == 1)
                           break;
                   fprintf(stderr, "%s: Bad buffer size '%s'\n", progname,
                           optarg);
                   ec = 2;
                   break;
           case 'c':       /* Count, # messages */
                   if(sscanf(optarg, "%ld", &NumMessages) == 1)
                           break;
                   fprintf(stderr, "%s: Bad message count '%s'\n",
                           progname, optarg);
                   ec = 2;
                   break;
           case 'C':       /* Don't check message contents */
                   ContentCheck = 0;
                   break;
           case 'd':       /* Print progress as a dot */
                   ProgDot = 1;
                   break;
           case 'D':       /* Dump contents of received message */
                   if(sscanf(optarg, "%ld", &DumpInMessage) == 1)
                           break;
                   fprintf(stderr,
                           "%s: Bad number of bytes to dump '-D %s'\n",
                           progname, optarg);
                   ec = 2;
                   break;
           case 'e':       /* Echo option */
                   echo = 1;
                   break;
           case 'E':       /* Echo Server/client */
                   EchoServerOption = optarg;
                   break;
           case 'f':       /* Flush read queue */
                   flush = 1;
                   break;
           case 'F':       /* Flow control message */
                   echo = 1;
                   if(sscanf(optarg, "%d", &FlowInterval) == 1 &&
                      FlowInterval >= 0)
                           break;
                   fprintf(stderr,
                           "%s: Bad Flow Control interval '-F %s'\n",
                           progname, optarg);
                   ec = 2;
                   break;
           case 'G':       /* Get timeout */
                   if(sscanf(optarg, "%ld", &play_l) == 1) {
                           GetTimeout = (MQLONG)play_l;
                           break;
                   }
                   fprintf(stderr, "%s: Bad Get timeout '%s'\n",
                           progname, optarg);
                   ec = 2;
                   break;
           case 'H':       /* Header information display */
                   if(sscanf(optarg, "%ld", &HdrDump) == 1 &&
                      HdrDump >= 0 && HdrDump <= HDR_DUMP_MAX)
                           break;

                   fprintf(stderr, "%s: Bad header dump level '%s'\n",
                           progname, optarg);
                   ec = 2;
                   break;
           case 'I':       /* Set Correlation Id */
                   CorrelId = optarg;
                   if(strlen(optarg) <= MQ_CORREL_ID_LENGTH)
                           break;
                   fprintf(stderr, "%s: Correlation Id too long '%s'\n",
                           progname, optarg);
                   ec = 2;
                   break;
           case 'l':       /* Learn Seed & NumMessages from 1st message */
                   SetParamsFromFirstMessage = 1;
                   break;
           case 'L':       /* Ignore 'Last' messages */
                   if(sscanf(optarg, "%ld", &IgnoreLast) == 1 &&
                      (IgnoreLast > 0 || IgnoreLast == -1))
                           break;
                   fprintf(stderr, "%s: Bad -L argument '%s'\n",
                           progname, optarg);
                   ec = 2;
                   break;
           case 'M':       /* Don't check message size */
                   DontCheckSize = 1;
                   break;
           case 'n':       /* No wait mode */
                   NoWait = 1;
                   break;
           case 'N':       /* Sequence number start */
                           /* Subtract 1 as we start at 1 anyway */
                   if(sscanf(optarg, "%ld", &SeqStart) == 1) {
                           SeqStart--;
                           break;
                   }
                   fprintf(stderr, "%s: Bad sequence number start '%s'\n",
                           progname, optarg);
                   ec = 2;
                   break;
           case 'o':       /* Message descriptor report option */
                   ec |= ParseReportOption(optarg);
                   break;
           case 'p':       /* Progress Messages */
                   if(sscanf(optarg, "%d", &ProgCnt) == 1)
                           break;
                   fprintf(stderr, "%s: Bad Progress count '%s'\n",
                           progname, optarg);
                   ec = 2;
                   break;
           case 'P':       /* Message priority */
                   PrioritySet = 1;
                   if(sscanf(optarg, "%d", &MsgPriority) == 1)
                           break;
                   fprintf(stderr, "%s: Bad message priority '%s'\n",
                           progname, optarg);
                   ec = 2;
                   break;
           case 'r':       /* Read messages */
                   ReadFromQueue = 1;
                   break;
           case 'R':       /* Random sized messages */
                   RandSize = 1;
                   break;
           case 's':       /* Stats every n messages */
                   if(sscanf(optarg, "%ld", &Stats) == 1)
                           break;
                   fprintf(stderr, "%s: Bad stats summary '%s'\n",
                           progname, optarg);
                   ec = 2;
                   break;
           case 'S':       /* Set Seed for random numbers */
                   if(sscanf(optarg, "%u", &seed) == 1) {
                           SeedGiven = 1;
                           break;
                   }
                   fprintf(stderr, "%s: Bad random number seed '%s'\n",
                           progname, optarg);
                   ec = 2;
                   break;
           case 't':       /* Trace level */
                   if(sscanf(optarg, "%d", &Trace) == 1)
                           break;
                   fprintf(stderr, "%s: Bad trace level '%s'\n", progname,
                           optarg);
                   ec = 2;
                   break;
           case 'u':       /* Operations in a Unit of Work */
                   if(sscanf(optarg, "%d", &UnitOfWork) == 1)
                           break;
                   fprintf(stderr, "%s: Bad unit of work number '%s'\n",
                           progname, optarg);
                   ec = 2;
                   break;
           case 'U':       /* Fail ratio of Unit of Work */
                   if(sscanf(optarg, "%d", &UnitsBackedOut) == 1)
                           break;
                   fprintf(stderr,
                           "%s: Bad fail ratio of unit-of-work number '%s'\n",
                           progname, optarg);
                   ec = 2;
                   break;
           case 'w':       /* Echo mode Write Wait */
                   if(sscanf(optarg, "%d", &WriteWait) == 1)
                           break;
                   fprintf(stderr, "%s: Bad arg to -w '%s'\n", progname,
                           optarg);
                   ec = 2;
                   break;
           case 'W':       /* Wait n seconds before each iteration */
                   if(sscanf(optarg, "%d", &LoopWait) == 1 && LoopWait >= 0)
                           break;
                   fprintf(stderr, "%s: Bad arg to -W '%s'\n", progname,
                           optarg);
                   ec = 2;
                   break;
           case 'x':       /* Help */
                   Usage(ec);
                   /* NOTREACHED */
      case 'X':   /* Don't check the X-or CRC in the header */
         CheckCRC = 0;
         break;
           case '?':       /* Whoops */
                   fprintf(stderr, "Unknown option\n");
                   ec = 2;
                   break;
           }
   }

   /* Echo Server option -- -E
    * It isn't 'till now that we know how to interpret it.
    * If we are the server (reads from queues) if should be >= 1,
    * the client name of the queue shouldn't be too long.
    */
   EchoServer = 0;
   if(EchoServerOption != (char*)NULL) {
           if(ReadFromQueue) {
                   EchoServer = 1;
                   if((sscanf(EchoServerOption, "%d",
                                   &EchoQueuesOpen) != 1 ||
                           EchoQueuesOpen <= 0)) {
                           fprintf(stderr,
                                   "%s: Bad # echo queues open '%s'\n",
                                   progname, EchoServerOption);
                           ec = 2;
                   }
           }
           else
                   if(strlen(EchoServerOption) >= MQ_Q_NAME_LENGTH) {
                           fprintf(stderr,
                                   "%s: Echo queue name too long: -E %s\n",
                                   progname, EchoServerOption);
                           ec = 2;
                   }
   }

   /* Return if no Queue specified.
    * NB If we are echoing we need the name of the echo queue as well,
    * unless we are an echo server in which case we get the queue from
    * the message.
    */
   if(argc != optind + 2 + (EchoServer ? 0 : echo)) {
           fprintf(stderr, "%s: Wrong number of arguments\n", progname);
           ec = 1;
   }

   /* Some options don't go well with others */
   if(ReadFromQueue && WriteWait) {
           fprintf(stderr, "%s: Can't have -r with -w option\n",
                   progname);
           ec = 1;
   }

   /* Or without others */
   if(SetParamsFromFirstMessage && !ReadFromQueue) {
           fprintf(stderr, "%s: Can't have -l without -r option\n",
                   progname);
           ec = 1;
   }

   /* NB MsgPriority of -1 on read means send the same priority back */
   if(PrioritySet &&
      (MsgPriority > 9 || MsgPriority < (ReadFromQueue ? -1 : 0))) {
           fprintf(stderr, "%s: message priority out of range: %d\n",
                   progname, MsgPriority);
           ec = 1;
   }

   if(ec)  /* Something go wrong ? */
           Usage(ec);

   /* Get arguments */
   QueueManager = argv[optind];
   QueueName = argv[optind + 1];
   if(echo && !EchoServer)
           EchoQueueName = argv[optind + 2];

   if(Trace) {
           printf("%s: Trace level %d ", progname, Trace);
           printf("ReadFromQueue=%d Echo=%d Flush=%d NoWait=%d\n",
                   ReadFromQueue, echo, flush, NoWait);
           printf("BufferSize=%d Random=%d Seed %d Sequence start %ld\n",
                   BufferSize, RandSize, seed, SeqStart + 1);
           printf("LoopWait=%d WriteWait=%d IgnoreLast=%d ", LoopWait,
                   WriteWait, IgnoreLast);
           printf("SetParamsFromFirstMessage=%d\n",
                   SetParamsFromFirstMessage);
           printf("HdrDump=%d GetTimeout=%d FlowInterval=%d ContentCheck=%d\n",
                   HdrDump, GetTimeout, FlowInterval, ContentCheck);

           printf("QueueManager '%s' QueueName '%s' EchoQueueName '%s'\n",
                   QueueManager, QueueName,
                   EchoQueueName == (char*)NULL ?  "(unset)" :
                                                   EchoQueueName);

           printf("CorrelId '%s' ",
                   CorrelId == (char*)NULL ? "(unset)" : CorrelId);
           printf(PrioritySet ? "Priority %d " : "Priority unset",
                   MsgPriority);
           putchar('\n');
   }
}

/* Print usage message & exit.
 */
void    Usage(int ec)
{
   printf("Quick throughput test for a message queue\n");
   printf("Usage: %s [-options] QueueManager QueueName [EchoQueueName]\n",
                           progname);
   printf("Send messages to QueueName handled by QueueManager\n");
   printf("Options:\n");
   printf("-b n\tBuffer size (default %dK)\n", DefBufferSize / Kilo);
   printf("-c n\tCount, number of messages (default %d)\n",
                           DefNumMessages);
   printf("\tif -1 go until get interrupt\n");
   printf("-C\tdon't Check message Content\n");
   printf("-d\tprint a Dot as a progress message, default is wordy\n");
   printf("-D nn\tDump nn bytes of each message received\n");
   printf("-e\tEcho. Send message to QueueName, read reply from EchoQueueName\n");
   printf("\tIf -r read from QueueName, reply to EchoQueueName\n");
   printf("-E nn\tEchoing (-r), # reply queues to keep open\n");
   printf("\tReply to queue name in the message - server\n");
   printf("-E nam\tEchoing (no -r), reply_to_queue name put in the message\n");
   printf("\tRead reply from EchoQueueName - sender/client\n");
   printf("-f\tFlush the read queue before starting (echo mode)\n");
   printf("-F nn\tlike -e, read/write flow control msg every nn msgs\n");
   printf("-G n\ttimeout message Gets in n milliseconds\n");
   printf("-H n\tdisplay received message Header information\n");
   printf("-I txt\tuse a Correlation Identifier of txt, max 24 chars\n");
   printf("-l\tLearn Seed & NumMessages from 1st message\n");
   printf("-L n\tignore n Last messages - for closedown, -1 to ignore all\n");
   printf("-M\tdon't check Message size\n");
   printf("-n\tNo wait, show time spent waiting for a message to arrive\n");
   printf("-N n\tsequence Number of first message to send/expect\n");
   printf("-o str\tmessage descriptor report Option\n");
   printf("-p n\tProgress print every n messages, if trace >= %d",
                           ProgLev);
   printf(" (default %d)\n", DefProgCnt);
   printf("-P n\tPriority to put msgs to queue 0..9.");
   printf(" On read & n is -1 copy priority\n");
   printf("-r\tRead from the queue QueueName - default is to write\n");
   printf("-R\tRandom sized messages\n");
   printf("-s n\tprint a stats Summary every n messages, if trace >= %d\n",
                           ProgLev);
   printf("-S n\tSet Seed for random numbers\n");
   printf("-t n\tTrace level n, 0..9 - default %d\n", DefTrace);
   printf("-u n\tUnit of work ends every n operations\n");
   printf("-U n\tFail (backout) every n'th unit of work\n");
   printf("-w n\tWait n messages before reading (echo mode writer)");
   printf(" default 0\n");
   printf("-W n\tWait n seconds before each iteration\n");
   printf("-x\teXplain\n");
   printf("-X\tdon't check the X-or CRC in the header\n");
   printf("Version %s\n", "1.19.1.1 00/11/13 16:36:47");                /* SCCS number here */

   exit(ec);
}

/* Parse the Message Descriptor Report Option.
 * Return non zero on error (after wingeing).
 */
int     ParseReportOption(char* opt)
{
   Repo*   rp;

   /* Recognise this specially as we use it to set it to zero */
   if(strcmp(opt, "MQRO_NONE") == 0) {
           ReportOption = (MQLONG)MQRO_NONE;
           return(0);
   }

   /* Hunt through the list of options */
   for(rp = RepoList; rp->ro_name != (char*)NULL; rp++)
           if(strcmp(rp->ro_name, opt) == 0) {
                   ReportOption |= rp->ro_val;
                   return(0);
           }

   /* Whoops, someone doesn't know what they are doing */
   fprintf(stderr,
           "%s: Error: unknown Message Descriptor Report Option '%s'\n",
           progname, opt);

   return(1);
}

/* **** Initialisation ****
 *
 * Open a queue, WriteFlag is true if we want to write to the queue.
 * *hObjp has the queue handle pointer assigned there.
 * If EchoQueue open the echo queue.
 * Return QOpened the name of the queue that we opened.
 *        NULL on error.
 */
char*   OpenQueue(int WriteFlag, MQHOBJ* hObjp, int EchoQueue)
{
   MQLONG  CompCode, Reason;
   MQOD    ObjDesc         = { MQOD_DEFAULT };
   MQMD    Mqmd            = { MQMD_DEFAULT };
   MQPMO   Pmo             = { MQPMO_DEFAULT };
   MQLONG  flQueueFlags    = { 0 };
   char*   QOpened;

   QOpened = EchoQueue ? EchoQueueName : QueueName;

   if(Trace > 3)
           TimeMessage("About to open the queue: %s", QOpened);

   /* Open the Queue
    */
   strncpy(ObjDesc.ObjectName, QOpened, sizeof(ObjDesc.ObjectName));

   flQueueFlags |= WriteFlag ? MQOO_OUTPUT : MQOO_INPUT_SHARED;

   MQOPEN(hQm, &ObjDesc, flQueueFlags, hObjp, &CompCode, &Reason);

   if(Trace || CompCode == MQCC_FAILED)
           TimeMessage("MQOPEN %s Compcode(%ld) Reason(%ld)", QOpened,
                   CompCode, Reason);

   return((exit_c = CompCode == MQCC_FAILED) ? (char*)NULL : QOpened);
}

/* Find the echo queue that is named in the message options structure that
 * we got on the MQGET of the current message. Assign to *EMqmd.
 * The main problem that this routine faces is that it must only have at
 * most EchoQueuesOpen Queues open at any one time.
 * We also try to be intelligent (probably a bad thing to start with) and if
 * we have to close a queue: close the least recently used one.
 *
 * Return the name of the queue if we open it, NB this isn't NUL terminated,
 * it is in a MQCHAR48 buffer.
 * Return NULL if we can't open the queue - setting exit_c.
 */
char*   OpenEchoQueue(MQMD* Mqmd, MQHOBJ* obj)
{
   MQLONG  CompCode, Reason;
   MQLONG  flQueueFlags    = { MQOO_OUTPUT };
   MQOD    ObjDesc         = { MQOD_DEFAULT };
   Ecb*    eprobe;         /* Probe into echo queue buffers */
   Ecb*    eoldest;        /* Oldest one found */
                           /* Big enough to NUL terminate */
   char    qname[sizeof(eprobe->ec_qname) + 1];

   strncpy(qname, Mqmd->ReplyToQ, sizeof(qname) - 1);
   qname[sizeof(qname) - 1] = '\0';

   if(Trace > 5)
           TimeMessage("OpenEchoQueue: '%s'", qname);

   eoldest = EchoHandles;

   /* Scan to see if we know about it.
    * Note the oldest as we pass by.
    */
   for(eprobe = EchoHandles; eprobe < EchoHandles + EchoQueuesOpen;
                                                           eprobe++) {
           if(!strncmp(eprobe->ec_qname, Mqmd->ReplyToQ, MQ_Q_NAME_LENGTH)) {
                   /* Got it */
                   memcpy((void*)obj, (void*)&eprobe->ec_qmd,
                                                   sizeof(MQHOBJ));
                   if(Trace > 6)
                           printf("Q already open, last used %ldms ago\n",
                               MsDiff(ProgStart, now) - eprobe->ec_used);
                   eprobe->ec_used = MsDiff(ProgStart, now);
                   return(eprobe->ec_qname);
           }

           /* Older than the current one ? */
           if(eprobe->ec_used < eoldest->ec_used)
                   eoldest = eprobe;
   }

   /* We need to open a new queue.
    * Do we need to close the slot where we will store it ?
    */
   if(eoldest->ec_used != 0) {
           if(Trace > 6)
                   printf("Closing Q '%.*s'\n", MQ_Q_NAME_LENGTH,
                           eoldest->ec_qname);
           MQCLOSE(hQm, &eoldest->ec_qmd, MQCO_NONE, &CompCode, &Reason);
           eoldest->ec_used = 0;
   }

   if(Trace > 3)
           TimeMessage("OpenEchoQueue: opening '%s'", qname);

   strncpy(ObjDesc.ObjectName, Mqmd->ReplyToQ, sizeof(ObjDesc.ObjectName));

   MQOPEN(hQm, &ObjDesc, flQueueFlags, &eoldest->ec_qmd, &CompCode,
           &Reason);

   if(Trace > 2 || CompCode == MQCC_FAILED)
           TimeMessage("MQOPEN %s Compcode(%ld) Reason(%ld)", qname,
                   CompCode, Reason);

   if(CompCode == MQCC_FAILED) {
           exit_c = 2;
           return((char*)NULL);    /* Can't open queue */
   }

   /* New queue opened.
    * Copy information back to our caller, give it a name, & update
    * status into - when last used.
    */
   memcpy((void*)obj, (void*)&eoldest->ec_qmd, sizeof(MQHOBJ));
   strncpy(eoldest->ec_qname, qname, sizeof(eoldest->ec_qname));
   eoldest->ec_used = MsDiff(ProgStart, now);

   return(eoldest->ec_qname);
}

/* Flush the echo queue.
 * Achieve this by reading messages 'till there aren't any more.
 *
 * Return 0 on OK, true on error.
 */
int     FlushEchoQueue(void)
{
   MQLONG  CompCode, Reason;
   MQHOBJ  FhObj           =   MQHO_UNUSABLE_HOBJ;
   MQGMO   FGmo            = { MQGMO_DEFAULT };
   MQMD    FMqmd           = { MQMD_DEFAULT };
   MQLONG  MsgLen;
   char*   FQname;

   /* Open it for reading */
   if((FQname = OpenQueue(0, &FhObj, 1)) == (char*)NULL)
           return(1);

   /* Race around pulling messages off the queue.
    * Don't set the wait flag.
    */
   for(;;) {
           memcpy(FMqmd.MsgId   , MQMI_NONE, sizeof(FMqmd.MsgId));
           memcpy(FMqmd.CorrelId, MQCI_NONE, sizeof(FMqmd.CorrelId));

           if(Trace >= 6)
                   TimeMessage("About to MQGET (flush) on %s", FQname);
           MQGET(hQm, FhObj, &FMqmd, &FGmo, BufferSize, Message, &MsgLen,
                   &CompCode, &Reason);
           if(CompCode) {
                   TimeMessage("MQGET %s (flush) Compcode(%ld) Reason(%ld)",
                                   FQname, CompCode, Reason);
                   if(Reason == MQRC_NO_MSG_AVAILABLE)
                           break;  /* Nothing there */

         /* Extra brackets for tedious compilers */
                   if((exit_c = (CompCode == MQCC_FAILED)) != 0)
                           return(1);
           }
   }

   /* This is the only time that we actually bother to close a queue */
   MQCLOSE(hQm, &FhObj, MQCO_NONE, &CompCode, &Reason);

   if(Trace || CompCode == MQCC_FAILED)
           TimeMessage("MQCLOSE %s (flush) Compcode(%ld) Reason(%ld)",
                   FQname, CompCode, Reason);

   return(exit_c = CompCode == MQCC_FAILED);
}

/* Allocate the buffer.
 * Fill with a repeating string, just asterisks was too boring.
 * Add room for a NUL byte followed by the CRC, this is so that we don't
 * corrupt memory if we try to recalculate a CRC on a received message that
 * doesn't contain a NUL byte - as it should.
 */
void    Allocate(void)
{
   char*   m;
   size_t  ms;
   char*   s1;

   /* Ensure big enough to contain sequence #s, etc */
   ms = BufferSize > MIN_MESSAGE_SIZE ? BufferSize :
           MIN_MESSAGE_SIZE;

   if((Message = (char*)malloc(ms + CRC_BYTES + 1)) == (char*)NULL) {
           fprintf(stderr,
                   "%s: Error, could not allocate buffer of %d bytes\n",
                   progname, (int)ms);
           exit(4);
   }

   Message[ms] = '\0';

   /* Will with the repeating pattern */
   for(m = Message, s1 = MsgTxt; m < Message + ms;) {
           if(s1 >= MsgTxt + sizeof(MsgTxt) - 1)
                   s1 = MsgTxt;
           *m++ = *s1++;
   }
}

/* **** This is where the real work is done ****
 *
 * Send the messages.
 */
void      DoWrites(void)
{
   tval_t  TimeOfFirstMessage;             /* Start time */
   tval_t  StartOfWaitTime;                /* When start waiting on read */
   long    Duration;                       /* How long it takes to send */
   int     Iterations;
   int     SentLast = 0;                   /* True if sent 'L' message */
   MQLONG  PMsgLen;                        /* Size of message to Put */
   MQLONG  GMsgLen;                        /* Size of message to Get */
   MQLONG  EGMsgLen;                       /* Expected size of message to Get */
   MQLONG  CompCode;
   MQLONG  Reason;
   MQMD    Mqmd            = { MQMD_DEFAULT };
   MQHOBJ  hObj            =   MQHO_UNUSABLE_HOBJ;
   MQPMO   Pmo             = { MQPMO_DEFAULT };
   char*   WQname;                         /* Write queue */

   MQHOBJ  EhObj           =   MQHO_UNUSABLE_HOBJ; /* For the echo Q */
   MQGMO   EGmo            = { MQGMO_DEFAULT };
   MQMD    EMqmd           = { MQMD_DEFAULT };
   char*   EQname;                         /* Echo queue */

   /* Open the queues.
    * Echo queue in echo mode, possibly flushing it first.
    */
   if((WQname = OpenQueue(1, &hObj, 0)) == (char*)NULL)
           return;

   if(echo && flush && FlushEchoQueue())
           return;

   if(echo && (EQname = OpenQueue(0, &EhObj, echo)) == (char*)NULL)
           return;

   if(Trace) {
           TimeMessage("Start of putting %ld messages", NumMessages);
           TimeOfFirstMessage = now;
   }

   /* Do puts & gets take place in a Unit of Work ? */
   Pmo.Options  |= UnitOfWork ? MQPMO_SYNCPOINT : MQPMO_NO_SYNCPOINT;
   EGmo.Options |= UnitOfWork ? MQGMO_SYNCPOINT : MQGMO_NO_SYNCPOINT;

   /* Send priority for messages ? */
   if(PrioritySet)
           Mqmd.Priority = MsgPriority;

   /* Specify that it is text than can be converted */
   memcpy(Mqmd.Format, mqfmt_string_array, sizeof(Mqmd.Format));

   for(Iterations = 0; Iterations < NumMessages ||
                                   NumMessages == -1; ) {
           /* To sleep -- -W option ? */
           if(LoopWait > 0 && Iterations != 0)
                   SleepABit();

           /* NB 'L' in the first byte means last message */
           SetNow();
           /* Insert the message #, we don't actually count it 'till
            * we have put it.
            */
           sprintf(Message, "%c %d %ld %d %s",
                   Iterations + 1 == NumMessages ? 'L' : 'M',
                   Iterations + 1 + SeqStart, NumMessages, seed,
                   FormatNow());

           /* If interrupted, stop */
           if(GotSig) {
                   Message[0] = 'L';
                   SentLast = 1;
           }

           /* Do we have to say where to send the reply back to ? */
           if(EchoServerOption != (char*)NULL)
                   strncpy(Mqmd.ReplyToQ, EchoServerOption,
                           MQ_Q_NAME_LENGTH);

           /* Correlation Id ? */
           memcpy(Mqmd.CorrelId, MQCI_NONE, sizeof(Mqmd.CorrelId));
           if(CorrelId != (char*)NULL)
                   strncpy((char *)Mqmd.CorrelId, CorrelId, sizeof(Mqmd.CorrelId));

           PMsgLen = MsgSize();
           Mqmd.Report = ReportOption;

      InsertCRC(Message);

           if(Trace >= 6)
                   TimeMessage("About to MQPUT on '%s' '%.30s' len %ld",
                           WQname, Message, (long)PMsgLen);
           MQPUT(hQm, hObj, &Mqmd, &Pmo, PMsgLen, Message, &CompCode,
                           &Reason);
           if(CompCode) {
                   printf("MQPUT %s Compcode(%ld) Reason(%ld)\n", WQname,
                           CompCode, Reason);
                   if((exit_c = (CompCode == MQCC_FAILED)) != 0)
                           break;
           }

           BytesInterval += PMsgLen;       /* Maintain stats */
           Iterations++;

           if(Trace >= TR_SHOW_CONTENTS)
                   TimeMessage("TX wrote Message %d, len %d, '%s'",
                           Iterations, PMsgLen, Message);

           /* If interrupted, stop */
           if((!echo && GotSig) || SentLast)
                   break;

           /* Read an echo'd message ?
            * This really ought to be in a generic read function.
            * It's a loop so that we can retry if we get a timeout.
            * If there is a lag (write wait) allow messages to build up.
            * We may only expect a reply every FlowInterval messages.
            */
           if(echo && Iterations >= WriteWait &&
              (FlowInterval == 0 || Iterations % FlowInterval == 0)) {
                   ZeroTime(&StartOfWaitTime);
                                   /* Will take anything ! */
TryAgain:               memcpy(EMqmd.MsgId   , MQMI_NONE, sizeof(EMqmd.MsgId));
                   memcpy(EMqmd.CorrelId, MQCI_NONE, sizeof(EMqmd.CorrelId));
                   if(CorrelId != (char*)NULL)
                           strncpy((char *)EMqmd.CorrelId, CorrelId,
                                   sizeof(EMqmd.CorrelId));

                   if((!NoWait || !IsZeroTime(StartOfWaitTime)) &&
                                   GetTimeout == 0) {
                           EGmo.Options = MQGMO_WAIT;
                           EGmo.WaitInterval = MQWI_UNLIMITED;
                   }
                   else if(GetTimeout != 0) {
                           EGmo.Options = MQGMO_WAIT;
                           EGmo.WaitInterval = GetTimeout;
                   }
                   else
                           EGmo = DefGmo;  /* Reset to no wait */

                   if(Trace >= 6)
                           TimeMessage("About to MQGET on %s", EQname);
                   MQGET(hQm, EhObj, &EMqmd, &EGmo, BufferSize, Message,
                           &GMsgLen, &CompCode, &Reason);
                   if(CompCode) {
                           /* If we are in NoWait mode & there wasn't a
                            * message, print what has happened & retry.
                            */
                           if(NoWait && IsZeroTime(StartOfWaitTime) &&
                                       Reason == MQRC_NO_MSG_AVAILABLE) {
                                   if(Trace >= 3)
                                           TimeMessage("No message in echo queue - waiting");
                                   else
                                           SetNow();
                                   StartOfWaitTime = now;
                                   goto TryAgain;
                           }

                           TimeMessage("MQGET %s Compcode(%ld) Reason(%ld)",
                                           EQname, CompCode, Reason);

                           if(Reason == MQRC_NO_MSG_AVAILABLE)
                                   printf("MQGET: No Message Available%s\n",
                                           (GetTimeout != 0 ? " - timeout" : ""));

                      if((exit_c = (CompCode == MQCC_FAILED)) != 0)
                                   break;
                   }

                   if(Trace >= 6)
                           TimeMessage("MQGET OK on %s '%s' len %ld", EQname,
                                   Message, (long)GMsgLen);

                   /* Did we spend some time waiting for the message ? */
                   if(!IsZeroTime(StartOfWaitTime)) {
                           SetNow();
                           Duration = MsDiff(StartOfWaitTime, now);
                           WaitTime += Duration;
                           if(Trace >= 3)
                                   TimeMessage("Wait on read of %ldms",
                                           Duration);
                   }

                   /* Check that the size of message received was what we
                    * expected. Don't if random sizes, we don't know what
                    * we will be getting if we don't read back straight
                    * after we have sent it.
                    * This isn't an immediately fatal error.
                    */
                   EGMsgLen = ExpectedMsgSize();
                   if(!DontCheckSize && !(RandSize && WriteWait) &&
                                                   EGMsgLen != GMsgLen) {
                           TimeMessage("MQGET got msg len %ld, expected %ld",
                                   (long)GMsgLen, (long)EGMsgLen);
                           exit_c = 2;
                   }

                   CheckContents(Message, GMsgLen, Iterations - WriteWait,
                           "Message Echoed");

                   /* Show what we got ? */
                   if(HdrDump)
                           DumpHeader(&EMqmd);

                   if(DumpInMessage)
                           DumpText(Message, GMsgLen);

                   /* Is the other side closing ?
                    * Hmmm, this will leave messages in the queue if
                    * WriteWait != 0.
                    * If we are ignoring 'Lasts' - is this the last 'Last'
                    * that we must ignore - the English syntax is as bad
                    * as the C.
                    */
                   if(ContentCheck && Message[0] == 'L' &&
                      (IgnoreLast == 0 ||
                       (IgnoreLast > 0 && --IgnoreLast == 0))) {
                           if(Trace >= 2)
                                   TimeMessage("Partner closing");
                           break;
                   }
           }

           CheckUnitOfWork();

           /* This needs to be at the bottom of the loop to keep timing
            * right - ie in echo mode don't start 'till we have had an
            * echo. Even then this isn't quite right, but it is good enough
            * - especially over more than a few messages.
            */
           PrintTrace("TX", "Iteration %d of %d len %d", Iterations,
                   NumMessages, PMsgLen);
   }

   if(Trace) {
           TimeMessage("\nEnd of putting %ld messages", Iterations);

           Duration = MsDiff(TimeOfFirstMessage, now) -
                   (TotTimeLoopWait + TimeLoopWait);
           printf("Total TX time %ld.%3.3ld, %s per second\n",
                   Duration / 1000, Duration % 1000,
                   PrintAve((float)Iterations * 1000, Duration));
   }
}

/* Read everything out of the queue.
 */
void    DoReads(void)
{
   tval_t  TimeOfFirstMessage;             /* Start time */
   tval_t  StartOfWaitTime;                /* When start waiting on read */
   long    Duration;                       /* How long it takes to send */
   int     Iterations;                     /* Msgs actually got */
   int     SentLast = 0;                   /* True if sent 'L' message */
   long    ExpMsgLen;                      /* Expected message length */
   MQLONG  CompCode, Reason;
   MQHOBJ  hObj            =   MQHO_UNUSABLE_HOBJ;
   MQGMO   Gmo             = { MQGMO_DEFAULT };
   MQMD    Mqmd            = { MQMD_DEFAULT };
   char*   RQname;                         /* Read queue */

   MQMD    EMqmd           = { MQMD_DEFAULT };     /* For the echo Q */
   MQHOBJ  EhObj           =   MQHO_UNUSABLE_HOBJ;
   MQPMO   EPmo            = { MQPMO_DEFAULT };
   MQLONG  MsgLen;
   char*   EQname;                         /* Echo queue */

   int     First           = 1;

   /* Open the queues.
    * Echo queue in echo mode, possibly flushing it first.
    */
   if((RQname = OpenQueue(0, &hObj, 0)) == (char*)NULL)
           return;

   /* If we are echoing to the name in the queue header, we do things
    * differently. If it is just one queue ensure that we can open it
    * now.
    */
   if(EchoServerOption == (char*)NULL &&
      echo && (EQname = OpenQueue(1, &EhObj, 1)) == (char*)NULL)
           return;

   /* Do puts & gets take place in a Unit of Work ? */
   EPmo.Options |= UnitOfWork ? MQPMO_SYNCPOINT : MQPMO_NO_SYNCPOINT;
   Gmo.Options  |= UnitOfWork ? MQGMO_SYNCPOINT : MQGMO_NO_SYNCPOINT;

   /* Receive however many messages that we must */
   for(Iterations = 0; NumMessages == -1 || Iterations < NumMessages; ) {
           ZeroTime(&StartOfWaitTime);

           /* To sleep -- -W option ? */
           if(LoopWait > 0 && Iterations != 0)
                   SleepABit();

           /* If it is text, convert to our CCSID/codeset */
           Gmo.Options = MQGMO_CONVERT;

TryAgain:       /* Will take anything ! */
           memcpy(Mqmd.MsgId   , MQMI_NONE, sizeof(Mqmd.MsgId));

           /* Correlation Id ? */
           memcpy(Mqmd.CorrelId, MQCI_NONE, sizeof(Mqmd.CorrelId));
           if(CorrelId != (char*)NULL)
                   strncpy((char *)Mqmd.CorrelId, CorrelId, sizeof(Mqmd.CorrelId));

           /* Always wait for the first message */
           if((!NoWait || !IsZeroTime(StartOfWaitTime) ||
                                   Iterations == 0) &&
                           GetTimeout == 0) {
                   Gmo.Options |= MQGMO_WAIT;
                   Gmo.WaitInterval = MQWI_UNLIMITED;
           }
           else if(GetTimeout != 0) {
                   Gmo.Options |= MQGMO_WAIT;
                   Gmo.WaitInterval = GetTimeout;
           }
           else
                   Gmo = DefGmo;   /* Reset to no wait */

           if(Trace >= 6)
                   TimeMessage("About to MQGET on %s", RQname);
           MQGET(hQm, hObj, &Mqmd, &Gmo, BufferSize, Message, &MsgLen,
                   &CompCode, &Reason);
           if(CompCode) {
                   /* If we are in NoWait mode & there wasn't a
                    * message, print what has happened & retry.
                    */
                   if(NoWait && IsZeroTime(StartOfWaitTime) &&
                                   Reason == MQRC_NO_MSG_AVAILABLE) {
                           if(Trace >= 3)
                                   TimeMessage("No message in read queue - waiting");
                           else
                                   SetNow();
                           StartOfWaitTime = now;

                           /* Unless we are to timeout completely, try
                            * again.
                            */
                           if(GetTimeout == 0)
                                   goto TryAgain;

                           TimeMessage("Aborting on read timeout");
                           exit_c = 2;
                           break;
                   }

                   TimeMessage("MQGET %s Compcode(%ld) Reason(%ld)",
                                   RQname, CompCode, Reason);

                   if(Reason == MQRC_NO_MSG_AVAILABLE)
                           printf("MQGET: No Message Available%s\n",
                                   (GetTimeout != 0 ? " - timeout" : ""));

                   if((exit_c = (CompCode == MQCC_FAILED)) != 0)
                           break;
           }

           Iterations++;           /* We read something */

           /* Show when we started ? */
           if(First && Trace) {
                   TimeMessage("Received first message");
                   TimeOfFirstMessage = now;
                   First = 0;
           }

           BytesInterval += MsgLen;        /* Maintain stats */

           PrintTrace("RX", "Iteration %d of %d len %d", Iterations,
                   NumMessages, MsgLen);
           if(Trace >= TR_SHOW_CONTENTS && !DumpInMessage)
                   printf("Msg %d len %d, text '%s'\n", Iterations,
                           MsgLen, Message);

           /* This has to be before we check the size, 'cos we may
            * well set the seed from that.
            */
           CheckContents(Message, MsgLen, Iterations, "Message Received");

           /* Check that the size of message received was what we
            * expected. Don't if random sizes, we don't know what
            * we will be getting.
            * This isn't an immediately fatal error.
            */
           ExpMsgLen = RandSize ? MsgSize() : BufferSize;
           if(!DontCheckSize && (!RandSize || SeedGiven) &&
                                                   MsgLen != ExpMsgLen) {
                   TimeMessage("MQGET got msg len %ld, expected %ld",
                           (long)MsgLen, ExpMsgLen);
                   exit_c = 2;
           }

           /* Did we spend some time waiting for the message ? */
           if(!IsZeroTime(StartOfWaitTime)) {
                   SetNow();
                   Duration = MsDiff(StartOfWaitTime, now);
                   WaitTime += Duration;
                   if(Trace >= 3)
                           TimeMessage("Wait on read of %ldms", Duration);
           }

           /* Show what we got ? */
           if(HdrDump)
                   DumpHeader(&Mqmd);

           if(DumpInMessage)
                   DumpText(Message, MsgLen);

           /* Is the other side closing ? */
           if(ContentCheck && Message[0] == 'L' &&
              (IgnoreLast == 0 ||
               (IgnoreLast > 0 && --IgnoreLast == 0))) {
                   if(Trace >= 2)
                           TimeMessage("Partner closing");
                   break;
           }

           /* Echo the message ?
            * Send back what we got. If we are doing random sizes
            * this will vary.
            * If there is a flow inteval only send one in FlowInterval,
            * note that unlike the sender we send a reply on the first
            * message.
            * Note that although we send an extra one to start with,
            * we don't send a reply to the last one, so overall it is
            * the right number of replies.
            */
           if(echo &&
              (FlowInterval == 0 || Iterations == 1 ||
               Iterations % FlowInterval == 0)) {
                   /* If interrupted, stop */
                   if(GotSig) {
                           Message[0] = 'L';
                           SentLast = 1;
                   }

         /* We need to recalculate the CRC as we may have
          * changed it by inserting a 'L'.
          * Don't do this if we don't know what the contents
          * of the message are.
          */
         if(ContentCheck)
            InsertCRC(Message);

                   /* Queue name in the header of the message that we
                    * have just received ?
                    */
                   if(EchoServerOption != (char*)NULL &&
                      (EQname = OpenEchoQueue(&Mqmd, &EhObj)) ==
                                                           (char*)NULL)
                           break;

                   /* Set priority - fixed or what it came in at ? */
                   if(PrioritySet)
                           EMqmd.Priority = MsgPriority == -1 ?
                                   Mqmd.Priority : MsgPriority;

                   /* Correlation Id ?
                    * Either copy what we find in the message that we
                    * just read, or what we have on the command line.
                    * The command line takes precedence.
                    */
                   memcpy(EMqmd.CorrelId, Mqmd.CorrelId,
                           sizeof(EMqmd.CorrelId));
                   if(CorrelId != (char*)NULL)
                           strncpy((char *)EMqmd.CorrelId, CorrelId,
                                   sizeof(EMqmd.CorrelId));

                   /* Copy format specification */
                   memcpy(EMqmd.Format, Mqmd.Format, sizeof(EMqmd.Format));
                   EMqmd.Report = ReportOption;

                   if(Trace >= 6)
                           TimeMessage("About to MQPUT on '%.*s' '%.30s'",
                                   MQ_Q_NAME_LENGTH, EQname, Message);
                   MQPUT(hQm, EhObj, &EMqmd, &EPmo, MsgLen, Message,
                           &CompCode, &Reason);
                   if(CompCode) {
                           printf("MQPUT %s Compcode(%ld) Reason(%ld)\n",
                                   EQname, CompCode, Reason);
                      if((exit_c = (CompCode == MQCC_FAILED)) != 0)
                                   break;
                   }
           }

           CheckUnitOfWork();

           /* If interrupted, stop */
           if((!echo && GotSig) || SentLast)
                   break;
   }

   /* Print stats ? */
   if(Trace) {
           TimeMessage("\nEnd of receiving %ld messages, got %d",
                   NumMessages, Iterations);

           Duration = MsDiff(TimeOfFirstMessage, now) -
                   (TotTimeLoopWait + TimeLoopWait);
           printf("Total RX time %ld.%3.3ld, %s per second\n",
                   Duration / 1000, Duration % 1000,
                   PrintAve((float)Iterations * 1000, Duration));
   }
}

/* Check to see if a unit of work needs to be completed.
 * We do so after UnitOfWork operations, if this is zero we don't bother
 * with units of work.
 * If UnitsBackedOut is non zero, back out every UnitsBackedOut'th unit of
 * work.
 * Since the semantics of MQCMIT and MQBACK are exactly the same I will
 * (just for a bit of fun) invoke them with a function pointer.
 * The MQENTRY is there for some PC compilers where the calling convention
 * used by some functions isn't ``default''; this is all about things like how
 * function arguments are passed, who tidies the stack, etc; horrible details
 * that you just don't want to know about - definitely non ansi.
 */
void    CheckUnitOfWork(void)
{
   MQLONG  CompCode, Reason;
   char*   WhatOp;                         /* Name of what we do */
   void (*MQENTRY MqFun)(MQHCONN, PMQLONG, PMQLONG); /* Func to do it */

   if(!UnitOfWork)                         /* No syncpoints ? */
           return;

   if(++UwCount != UnitOfWork)             /* Not yet ? */
           return;

   UwCount = 0;

   /* We either commit it or back it out.
    * Set up a function pointer to whatever, and also note it's name.
    */
   if(UnitsBackedOut != 0 && ++UwUnits == UnitsBackedOut) {
           UwUnits = 0;
           MqFun = MQBACK;
           WhatOp = "MQBACK";
   }
   else {
           MqFun = MQCMIT;
           WhatOp = "MQCMIT";
   }

   if(Trace >= 6)
           TimeMessage("About to %s", WhatOp);

   (*MqFun)(hQm, &CompCode, &Reason);      /* Do it */

   if(CompCode)    /* Something go wrong ? */
           printf("%s Compcode(%ld) Reason(%ld)\n",
                   WhatOp, CompCode, Reason);
}

/* Check the contents of the message in Mbuf, that is MsgLen bytes long.
 * The expected sequence # is sequence.
 * what is some descriptive text for an error message.
 * If the text is not as expected generate an error message & set the
 * error code, don't kill the program.
 */
void    CheckContents(char* Mbuf, long MsgLen, int sequence, char* what)
{
   static  int     CheckedFirstMessage;
   int     fseq;                   /* Sequence # found */
   long    lmsgs;                  /* # msgs */
   int     lseed;                  /* Seed # in message */
   int     AdjFlowI;               /* Adjustment on FlowInterval */
   char* endhp;         /* Pointer to the end of the header */
   int   crc;

   /* Not to check.
    * But we do if we have to get the run parameters from the first
    * message.
    */
   if(!ContentCheck &&
                   !(!CheckedFirstMessage && SetParamsFromFirstMessage))
           return;

   /* Do we need to adjust the sequence # expected because we only
    * receive one in FlowInterval ?
    * This is only if we are the writer.
    * Note that the first message has sequence # 1, not 0.
    */
   if(ReadFromQueue == 0 && FlowInterval != 0) {
           AdjFlowI = FlowInterval;
           if(!CheckedFirstMessage)
                   AdjFlowI--;
   }
   else
           AdjFlowI = 0;

   CheckedFirstMessage = 1;

   if(MsgLen < MIN_MESSAGE_SIZE)   /* Not enough to check */
           return;

   /* Check the CRC for the message header */
   if(CheckCRC) {
      crc = CalculateCRC(Mbuf);
      if((endhp = strchr(Mbuf, '\0'))[1] != HexDigits[crc & 0xf] ||
         endhp[2] != HexDigits[(crc >> 4) & 0xf]) {
         TimeMessage("Bad message CRC, sequence # %d in %s",
            sequence, what);
         exit_c = 2;
         return;
      }
   }

   /* Now check the rest of the message.
    * We don't care if this throws up an error, ie still try to extract
    * information from the header.
    */
   CheckMsgAlphaNumerics(Mbuf, MsgLen, sequence, what);

   if((Mbuf[0] != 'L' && Mbuf[0] != 'M') ||
       Mbuf[1] != ' ' ||
      sscanf(Mbuf + 2, "%d %ld %d", &fseq, &lmsgs, &lseed) != 3) {
           TimeMessage("Bad message text, sequence # %d in %s",
                   sequence, what);
           exit_c = 2;
           return;
   }

   /* OK ?
    * If this is the first message, maybe we ought to set the #
    * messages & the seed.
    */
   if(fseq == sequence + SeqStart - AdjFlowI) {
           if(SetParamsFromFirstMessage && sequence == 1) {
                   SeedGiven = 1;
                   seed = lseed;
                   srand(seed);
                   printf("%s: Got seed for random message size: %u\n",
                           progname, seed);
                   NumMessages = lmsgs;
           }
           return;
   }

   TimeMessage("Bad sequence in %s, got %d expected %d",
           what, fseq, sequence + SeqStart - AdjFlowI);
   exit_c = 2;
}

/* Check that the alphanumerics after the message header are as expected.
 */
void  CheckMsgAlphaNumerics(char* Mbuf, long MsgLen, int sequence, char* what)
{
   char* cp;      /* Where we look in the message */
   char* refp;    /* Reference pointer */

   if(MsgLen < MIN_MESSAGE_SIZE)
      return;

   /* Find the first byte that we should check.
    * NB that this will be immediately after the CRC bytes.
    */
   cp = strchr(Mbuf, '\0') + CRC_BYTES + 1;

   /* Sanity check */
   if(cp - Mbuf >= MsgLen) {
      TimeMessage(
"Error: message header longer than the message (%d > %d), sequence #%d in %s",
         cp - Mbuf, MsgLen, sequence, what);
      exit_c = 2;
      return;
   }

   /* Another one */
   if(cp - Mbuf >= sizeof(MsgTxt)) {
      TimeMessage("Message header too long: %d, sequence #%d in %s",
         cp - Mbuf, MsgLen, sequence, what);
      exit_c = 2;
      return;
   }

   MsgLen -= cp - Mbuf;
   refp = MsgTxt + (cp - Mbuf);
   while(--MsgLen >= 0) {
      if(refp - MsgTxt >= sizeof(MsgTxt) - 1)   /* Wrap round ? */
         refp = MsgTxt;

      if(*cp++ == *refp++)       /* OK ? */
         continue;

      TimeMessage("Message corrupted at offset %d, sequence #%d in %s",
         cp - Mbuf, sequence, what);
      exit_c = 2;
      return;
   }
}

/* **** Traceing and statistics routines ****
 *
 * Print Num1/Num2 as a floating point number.
 * Num2 is milliseconds.
 * Return the string.
 * Take care when Num2 is 0 - this is the reason for this otherwise
 * strange routine.
 */
char*   PrintAve(float Num1, long Num2)
{
   static  char    res[40];

   if(Num2 == 0)
           strcpy(res, "infinite");
   else
           sprintf(res, "%.3f", Num1 / Num2);

   return(res);
}

/* Set 'now' to the current time.
 * This is horribly operating system specific.
 */
void    SetNow(void)
{
#ifdef  AMQ_UNIX
   /* Sometimes takes one argument, sometimes 2 */
   gettimeofday((struct timeval *)&now, (struct timezone *)NULL);
#endif

#ifdef  AMQ_OS2
   /* The problem with OS/2 is that there isn't (as far as I can see)
    * a function which gives both the secs since the epoch & a sub
    * second value. OK, we do get everything from DosGetDateTime() but
    * it is all broken out into hh:mm:ss sort of thing, & getting back
    * to epoch seconds is non trivial.
    * What could happen is that in between the two calls a second ticks.
    * So we get the sub second value, seconds since the epoch & the
    * sub second value again. If the broken out second changes we get
    * the epoch seconds again - as we don't know which side of the
    * tick we were on when we got it.
    * This will be OK unless we somehow sleep for a long time part
    * way though this function.
    */
   time_t          ep_secs;        /* Secs since epoch */
   int             secs;           /* Seconds as per DosGetDateTime */
   DATETIME        DateTime;

   (void) DosGetDateTime(&DateTime);
   secs = DateTime.seconds;
   ep_secs = time((time_t*)NULL);
   (void) DosGetDateTime(&DateTime);
   if(secs != DateTime.seconds)
           ep_secs = time((time_t*)NULL);

   now.tv_sec = (long)ep_secs;
   now.tv_usec = (long)DateTime.hundredths * 10000;
#endif
}

/* If -W is set we have to sleep a bit at the top of the loops.
 * Do this, and maintain TimeLoopWait as the sum of how long we spent
 * asleep in this interval, this is so that stats can be maintained.
 * NB: sleep() only sleeps for about how long you ask it to.
 */
void    SleepABit(void)
{
   tval_t  StartOfSleep;           /* When started sleeping */

   if(Trace >= 7)
           TimeMessage("Start of sleep");

   SetNow();
   StartOfSleep = now;
   sleep(LoopWait);
   SetNow();
   TimeLoopWait += MsDiff(StartOfSleep, now);

   if(Trace >= 7)
           TimeMessage("End of sleep");
}

/* Format the current time up in a nice way and return a pointer to
 * a static buffer containing the string.
 */
char*   FormatNow(void)
{
   struct  tm*     tp;
   static  char    FBuf[60];       /* What we return */

   tp = localtime((time_t *)&now.tv_sec);

   sprintf(FBuf, "%2.2d:%2.2d:%2.2d %ld.%3.3ld",
           tp->tm_hour, tp->tm_min, tp->tm_sec,
           (long)(now.tv_sec), now.tv_usec / 1000);

   return(FBuf);
}

/* Output a message, followed by the time.
 * Set now as a side effect.
 */
void    TimeMessage(char* fmt, ...)
{
   va_list         ap;

   SetNow();

   va_start(ap, fmt);
   vfprintf(stdout, fmt, ap);
   va_end(ap);

   printf(" at %s\n", FormatNow());
}

/* Print a trace message, if we are supposed to, or just a '.' or nothing
 * at all.
 * dirn is "TX" or "RX" as appropriate.
 * Allways set 'now' as a side effect.
 * ProgMsg controls if we are to print a progress message.
 */
void    PrintTrace(char* dirn, char* fmt, ...)
{
   long            Duration;       /* Of interval - milliseconds */
   struct  tm*     tp;
   va_list         ap;

   SetNow();

   if(MsgNo == 0)          /* First message ? */
           IntervalStart = now;

   ++MsgNo;
   ++ProgMsg;

   if(Trace < ProgLev)     /* Nothing at all ? */
           return;

   /* Nothing at all 'cos we haven't reached a count yet ? */
   if(!(Trace < ProgLev + 1 && ProgMsg < ProgCnt)) {
           /* Print something */
           ProgMsg = 0;

           /* Just a '.' ? */
           if(Trace == ProgLev && ProgDot)
                   printf(".");
           else {
                   va_start(ap, fmt);
                   vfprintf(stdout, fmt, ap);
                   va_end(ap);

                   tp = localtime((time_t *)&now.tv_sec);

                   printf(" at %2.2d:%2.2d:%2.2d %ld.%3.3ld\n",
                           tp->tm_hour, tp->tm_min, tp->tm_sec,
                           (long)now.tv_sec, now.tv_usec);
           }
           fflush(stdout);
   }

   /* Print an intermediate summary ? */
   if(Stats == 0 || MsgNo % Stats != 0)
           return;         /* No */

   tp = localtime((time_t *)&now.tv_sec);

   printf("\nSummary on msg# %d at %2.2d:%2.2d:%2.2d %ld.%3.3ld\n",
           MsgNo, tp->tm_hour, tp->tm_min, tp->tm_sec, (long)now.tv_sec,
           now.tv_usec / 1000);

   Duration = MsDiff(IntervalStart, now) - TimeLoopWait;
   printf("Interval %s time %.3f, %s msgs/second", dirn,
                   (float)Duration / 1000,
                   PrintAve((float)Stats * 1000, Duration));
   printf(" %.0f bytes (%.0f/second)\n", BytesInterval,
           BytesInterval * 1000 / Duration);

   if(NoWait)
           printf("Wait time this interval: %.3fs\n",
                   (float)WaitTime / 1000);

   if(TimeLoopWait)
           printf("Elapsed time %.3fs, time sleeping %.3fs\n",
                   (float)(Duration + TimeLoopWait) / 1000,
                   (float)TimeLoopWait / 1000);

   IntervalStart = now;

   /* Maintain stats */
   if(MaxInterval == -1 || Duration > MaxInterval)
           MaxInterval = Duration;
   if(MinInterval == -1 || Duration < MinInterval)
           MinInterval = Duration;
   TotInterval += Duration;
   TotMsgs += Stats;
   NumIntervals++;

   /* Wait time stats */
   if(MaxWaitTime == -1 || MaxWaitTime > WaitTime)
           MaxWaitTime = WaitTime;
   if(MinWaitTime == -1 || MinWaitTime < WaitTime)
           MinWaitTime = WaitTime;
   TotWaitTime += WaitTime;
   WaitTime = 0;
   TotTimeLoopWait += TimeLoopWait;
   TimeLoopWait = 0;

   /* What data volume ? */
   TotBytesSent += BytesInterval;
   BytesInterval = 0.0;
}

/* End of program, print intermediate stats - if recorded.
 * Print how long intervals took (seconds), and what the message
 * rate was.
 */
void    PrintStats(void)
{
   if(Stats == 0 || MinInterval == -1)
           return;

   printf("\nInterval duration for %ld messages min/avg/max =", Stats);
   printf(" %.3f/%s/%.3f\n", (float)MinInterval / 1000,
           PrintAve((float)TotInterval / 1000, NumIntervals),
           (float)MaxInterval / 1000);

   /* Take care: PrintAve returns a pointer to a static buffer */
   printf("Messages per second min/avg/max = %s/",
                   PrintAve((float)Stats * 1000, MaxInterval));
   printf("%s/", PrintAve((float)TotMsgs * 1000, TotInterval));
   printf("%s\n", PrintAve((float)Stats * 1000, MinInterval));

   printf("Data volume: %.0f bytes in %.2f seconds (%.0f/second)\n",
           TotBytesSent, (float)TotInterval / 1000,
           TotBytesSent / ((float)TotInterval / 1000));

   if(NoWait) {
           printf("Total Wait time this run: %.3fs",
                   (float)TotWaitTime / 1000);
           printf(" min/avg/max = %.3f/%.3f/%.3f\n",
                   (float)MaxWaitTime / 1000,
                   (float)TotWaitTime / NumIntervals / 1000,
                   (float)MinWaitTime / 1000);
   }

   if(TotTimeLoopWait)
           printf("Total time sleeping this run: %.3fs\n",
                   (float)TotTimeLoopWait / 1000);
}

/* Dump the information that we got with the message - header info.
 * How much we dump depends on the header dump level - HdrDump.
 * How and in which order we do this is *very* arbitrary.
 */
void    DumpHeader(MQMD* InQmd)
{
   printf("MsgId ");
   DumpStr((char *)InQmd->MsgId, sizeof(InQmd->MsgId));
   printf("\nCorrelId ");
   DumpStr((char *)InQmd->CorrelId, sizeof(InQmd->CorrelId));
   printf("\n");

   printf("Uid '%12.12s' Format '%8.8s'\n",InQmd->UserIdentifier,
                                                   InQmd->Format);
   printf("Priority %ld Persist %ld CCSID %ld Encode %ld Type %ld\n",
           (long)InQmd->Priority, (long)InQmd->Persistence,
           (long)InQmd->CodedCharSetId, (long)InQmd->Encoding,
           (long)InQmd->MsgType);

   if(HdrDump < 2)
           return;

   printf("ReplyToQMgr '%48.48s'\n", InQmd->ReplyToQMgr);
   printf("ReplyToQ    '%48.48s'\n", InQmd->ReplyToQ);
   printf("PutApplName '%28.28s' PutApplType %ld ApplOriginData '%4.4s'\n",
           InQmd->PutApplName, (long)InQmd->PutApplType,
           InQmd->ApplOriginData);

   if(HdrDump < 3)
           return;

   printf("PutDate '%8.8s' PutTime '%8.8s' Expiry ",
           InQmd->PutDate, InQmd->PutTime);

   if(InQmd->Expiry == MQEI_UNLIMITED)
           printf("unlimited");
   else
           printf("%ld.%ld secs", (long)InQmd->Expiry / 10,
                   (long)InQmd->Expiry % 10);

   printf(" Report %#ld\n", (long)InQmd->Report);

   printf("Feedback %ld BackoutCount %ld\n",
           (long)InQmd->Feedback, (long)InQmd->BackoutCount);

   if(HdrDump < 4)
           return;

   printf("AccountingToken  ");
   DumpStr((char *)InQmd->AccountingToken, sizeof(InQmd->AccountingToken));
   printf("\n");
   printf("ApplIdentityData ");
   DumpStr(InQmd->ApplIdentityData, sizeof(InQmd->ApplIdentityData));
   printf("\n");

   if(HdrDump < 5)
           return;

   printf("StrucId '%4.4s' Version %ld\n",
           InQmd->StrucId, (long)InQmd->Version);
}

/* Dump len bytes at str.
 * Since it is possible (probable ?) that some of the bytes may be control
 * characters or even stranger, we first dump as hex then as characters -
 * taking care to output funny characters.
 * Siglen
 */
void    DumpStr(char* str, int len)
{
   char*   cp;
   int     i;

   printf("#");

   /* First as hex */
   for(i = 0, cp = str; i < len; i++, cp++)
           printf("%2.2x", 0xff & *cp);

   printf(" '");

   /* Next as printable characters */
   for(i = 0, cp = str; i < len; i++, cp++)
           if(isascii(*cp) && isprint(*cp))
                   printf("%c", *cp);
           else
                   printf(" ");

   printf("'");
}

/* Display the contents of Message, there are Len bytes there.
 * Display at most DumpInMessage bytes, -1 means everything.
 */
void    DumpText(char* Message, MQLONG Len)
{
   long    offset;
   char*   cp;
   int     i;

   TimeMessage("%ld (%#lx) bytes received", (long)Len, (long)Len);

   if(DumpInMessage != -1 && DumpInMessage < Len)
           Len = DumpInMessage;

   /* Dump 16 bytes at a time, hex, then characters */
   for(offset = 0; Len > offset; offset += 16, Message += 16) {
           printf("%6.6ld #%4.4lx ", offset, offset);
           for(i = 0, cp = Message; i < 16; i++, cp++)
                   if(offset + i < Len)
                           printf("%2.2x ", 0xff & *cp);
                   else
                           printf("   ");
           printf(":");
           for(i = 0, cp = Message; i < 16; i++, cp++)
                   if(offset + i < Len && isascii(*cp) && isprint(*cp))
                           printf("%c", *cp);
                   else
                           printf(" ");
           printf(":\n");
   }
}

/* **** Misc ****
 *
 * Interrupt catching function.
 */
void    CatchIntr(int sig)
{
   GotSig = 1;
}

/* Calculate the milliseconds time difference between 2 times.
 */
long    MsDiff(tval_t start, tval_t end)
{
   long    res;

   res = (end.tv_sec - start.tv_sec) * 1000;

   res += (end.tv_usec - start.tv_usec) / 1000;

   return(res);
}

/* Set a time structure to zero.
 */
void    ZeroTime(tval_t* tp)
{
   tp->tv_sec = 0;
   tp->tv_usec = 0;
}

/* Return true if tp is zero.
 */
int     IsZeroTime(tval_t tp)
{
   return(tp.tv_sec == 0 && tp.tv_usec == 0);
}

/* Compute a random message size.
 * Make it no bigger than BufferSize.
 * If we are checking message contents the message needs to be big enough
 * to contain sequence #, etc, ie at least MIN_MESSAGE_SIZE.
 * We may not be checking messages, but ensure that the first message
 * is always big enough to contain up to the seed.
 *
 * RAND_MAX generally (ie exhaustive survey of 2 machines) seems to be 32K.
 * We need to do something if we are to generate bigger messages than this.
 */
long    RandMsgSize(void)
{
   static  int     DoneFirstMessage;
   unsigned long   size;
   size = rand();

   if(RAND_MAX < BufferSize)
           size = size * RAND_MAX + rand();

   size %= BufferSize;

   if((ContentCheck || !DoneFirstMessage) && size < MIN_MESSAGE_SIZE)
           size = MIN_MESSAGE_SIZE;

   DoneFirstMessage = 1;

   return(size);
}

/* Compute the size of a message to send.
 * Only use random numbers if we have to, store the size if there
 * is going to be other messages before we read it back.
 */
long    MsgSize(void)
{
   long    size;

   if(!RandSize)
           return(BufferSize);

   size = RandMsgSize();

   /* Record in circular buffer if there is a delay before we
    * read the message back again.
    */
   if(WriteWait && echo) {
           *WriteSize++ = size;
           if(WriteSize >= SizeBuff + WriteWait + 1)
                   WriteSize = SizeBuff;
   }

   LastMsgSize = size;

   return(size);
}

/* How long is the message that we are expecting to receive ?
 * The complication is if there have been other messages sent
 * since we sent the message that we are expecting.
 */
long    ExpectedMsgSize(void)
{
   long    l;

   if(!RandSize)
           return(BufferSize);

   if(!(WriteWait && echo))
           return(LastMsgSize);

   l = *ReadSize++;

   /* Gone round the circular buffer ? */
   if(ReadSize >= SizeBuff + WriteWait + 1)
           ReadSize = SizeBuff;

   return(l);
}

/* Insert a CRC bytes immediately after message header, ie after the NUL
 * byte which ends the sprintf'd string that contains the message sequence
 * number, etc.
 * The CRC is a simple exclusive OR, the purpose is to allow detection of
 * gross errors in header.
 * Assume that we can always insert the CRC as the minimum size of the
 * message buffer will allow room.
 * NB: The representation is least significant quartet first.
 */
void  InsertCRC(char* Msg)
{
   char* EndH;    /* End of header */
   int   crc;

   crc = CalculateCRC(Msg);
   (EndH = strchr(Msg, '\0'))[1] = HexDigits[crc & 0xf];
   EndH[2] = HexDigits[(crc >> 4) & 0xf];
}

/* Return an X-OR CRC of the string at Msg.
 */
char  CalculateCRC(char* Msg)
{
   char  crc;

   for(crc = '\0'; *Msg != '\0';)
      crc ^= *Msg++;

   return(crc);
}

/* Return a copy of str.
 * Bomb the program if no memory.
 */
char*   strmake(char* str)
{
   char*   cpy;

   if((cpy = (char*)malloc(strlen(str) + 1)) == (char*)NULL) {
           fprintf(stderr, "%s: Out of dynamic memory\n", progname);
           exit(2);
   }

   return(strcpy(cpy, str));
}

#ifdef AMQ_PC
/* This provides an implementation of getopt for OS/2 and Windows NT.
 * We can not use the common services function xugGetOpt() as it is
 * case insensitive. If this ever changes, then the code here should
 * be modified as appropriate
 *
 * HEADER FROM ORIGINAL FILE - SUPPLIED BY ALAIN WILLIAMS (A8DDW)
 * who saved it from the news: comp.sources.misc: Volume 2, Issue 90,
 * April 1988. Submitted-By: "Mike Patnode" <mpatnode@polyslo.UUCP>
 *
 * Here's something you've all been waiting for:  the AT&T public domain
 * source for getopt(3).  It is the code which was given out at the 1985
 * UNIFORUM conference in Dallas.  I obtained it by electronic mail
 * directly from AT&T.  The people there assure me that it is indeed
 * in the public domain.
 *
 * There is no manual page.  That is because the one they gave out at
 * UNIFORUM was slightly different from the current System V Release 2
 * manual page.  The difference apparently involved a note about the
 * famous rules 5 and 6, recommending using white space between an option
 * and its first argument, and not grouping options that have arguments.
 * Getopt itself is currently lenient about both of these things White
 * space is allowed, but not mandatory, and the last option in a group can
 * have an argument.  That particular version of the man page evidently
 * has no official existence, and my source at AT&T did not send a copy.
 * The current SVR2 man page reflects the actual behavor of this getopt.
 * However, I am not about to post a copy of anything licensed by AT&T.
 */
int getopt(int argc, char **argv, char *opts)
{
   static int sp = 1;
   register int c;
   register char *cp;

   if(sp == 1)
   {
           if(optind >= argc ||
              argv[optind][0] != '-' || argv[optind][1] == '\0')
           {
                   return(EOF);
           }
           else if(strcmp(argv[optind], "--") == 0)
           {
                   optind++;
                   return(EOF);
           }
   }

   optopt = c = argv[optind][sp];

   if(c == ':' || (cp=strchr(opts, c)) == NULL)
   {
           if(argv[optind][++sp] == '\0')
           {
                   optind++;
                   sp = 1;
           }
           return('?');
   }

   if(*++cp == ':')
   {
           if(argv[optind][sp+1] != '\0')
           {
                   optarg = &argv[optind++][sp+1];
           }
           else if(++optind >= argc)
           {
                   sp = 1;
                   return('?');
           }
           else
           {
                   optarg = argv[optind++];
           }

           sp = 1;
   }
   else
   {
           if(argv[optind][++sp] == '\0')
           {
                   sp = 1;
                   optind++;
           }
           optarg = NULL;
   }
   return(c);
}
#endif

/* end */
